-- get the addon namespace
local addon, ns = ...
-- get the config values
local cfg = ns.cfg
-- holder for some lib functions
local lib = _G["HEDD_lib"] or CreateFrame("Frame","HEDD_lib")
ns.lib = lib

cfg.aura_unit=nil
local aura_debug_var=false
--[[unitAuraUpdateInfo={
	isFullUpdate=boolean
	updatedAuraInstanceIDs={number}
	removedAuraInstanceIDs={number}
	addedAuras={UnitAuraInfo}
}

UnitAuraInfo={
applications	(	number	)	
auraInstanceID	(	number	)	
canApplyAura	(	boolean	)	Whether or not the player can apply this aura.
charges	(	number	)	
dispelName	(	string?	)	
duration	(	number	)	
expirationTime	(	number	)	
icon	(	number	)	
isBossAura	(	boolean	)	Whether or not this aura was applied by a boss.
isFromPlayerOrPlayerPet	(	boolean	)	Whether or not this aura was applied by the player or their pet.
isHarmful	(	boolean	)	Whether or not this aura is a debuff.
isHelpful	(	boolean	)	Whether or not this aura is a buff.
isNameplateOnly	(	boolean	)	Whether or not this aura should appear on nameplates.
isRaid	(	boolean	)	Whether or not this aura meets the conditions of the RAID aura filter.
isStealable	(	boolean	)	
maxCharges	(	number	)	
name	(	string	)	The name of the aura.
nameplateShowAll	(	boolean	)	Whether or not this aura should be shown on all nameplates, instead of just the personal one.
nameplateShowPersonal	(	boolean	)	
points	(	array	)	Variable returns - Some auras return additional values that typically correspond to something shown in the tooltip, such as the remaining strength of an absorption effect.
sourceUnit	(	string?	)	Token of the unit that applied the aura.
spellId	(	number	)	The spell ID of the aura.
timeMod	(	number	)	
}]]


--[[lib.AurasCount = function(spells,cast)
	cfg.aurabuff=1
	for spell,value in pairs(spells) do
		cfg.aurabuff=cfg.aurabuff * (lib.IsAura(spell,cast) and value or 1)
	end
	return cfg.aurabuff
end]]


local function aura_debug(aura,event,unitID,info)
	if aura_debug_var then
		unitID=unitID or cfg.aura_unit
		event=event or "Update"
		info=info or ""
		if aura then
			print(GetTime().." "..event.." on "..unitID.." "..aura.name.."("..aura.spellId..") "..aura.auraInstanceID.." "..info)
		else
			print(GetTime().." "..event.." on "..unitID.." "..info)
		end
	end
end

lib.AddAura = function(spell,spellId,atype,unit,sourceUnit,refresh)
	if not GetSpellInfo(spellId) then
		return
	end
		
	spellId = spellId or spell
	atype=atype or "buff"
	if atype=="buff" then
		unit=unit or "player"
	else
		unit=unit or "target"
	end

	refresh=refresh or 0

	cfg.watchunits[unit]=true
	cfg.aura_unit_id[unit] = cfg.aura_unit_id[unit] or {}
	cfg.aura_unit_instance[unit]=cfg.aura_unit_instance[unit] or {}
	
	cfg.aura[spell] = cfg.aura[spell] or {}
	cfg.aura[spellId] = cfg.aura[spell]
	cfg.aura_unit_id[unit][spellId] = cfg.aura[spell]
	cfg.aura[spell].spellId=spellId
	cfg.aura[spell].Name=spell
	cfg.aura[spell].spell=spell
	cfg.aura[spell].LocalName,_,cfg.aura[spell].icon = GetSpellInfo(spellId)
	--cfg.aura[spell].icon = GetSpellTexture(spellId)
	cfg.aura[spell].unit=unit
	cfg.aura[spell].sourceUnit=sourceUnit or "player"
	cfg.aura[spell].Type=atype
	cfg.aura[spell].func={}
	--cfg.aura[spell].events={}
	cfg.aura[spell].instance=nil
	lib.SetAuraRefresh(spell,refresh)
end

local AURA_EVENTS={
OnUpdate = true,
OnApply = true,
OnReApply = true,
OnRemove = true,
OnStacks = true
}

lib.SetAuraFunction = function(spell,event,func,run)
	if cfg.aura[spell] and AURA_EVENTS[event] then
		if run then
			func(spell)
		end
		cfg.aura[spell].func[event]=cfg.aura[spell].func[event] or {}
		table.insert(cfg.aura[spell].func[event],#cfg.aura[spell].func[event]+1,func)
		return #cfg.aura[spell].func[event]
	end
end

lib.RunAuraFunction = function(spell,event)
	if cfg.aura[spell].func[event] then
		for num_func,func in pairs(cfg.aura[spell].func[event] ) do
			func(cfg.aura[spell].spell,num_func)
		end
	end
end

local function CheckDispelAura(aura,unitID)
	unitID=unitID or cfg.aura_unit
	if UnitExists(unitID) and cfg.dispell[unitID] then
		if unitID=="target" and (UnitIsFriend("player", "target") or aura.isHarmful) then return false --and aura.isHarmful
		elseif unitID=="player" and aura.isHelpful then return false end
		
		aura.dispelName=aura.dispelName or "Other"
		if cfg.dispell[unitID].dispelName["Enrage"] then 
			if aura.dispelName=="" and aura.isStealable then aura.dispelName="Enrage" end
			if hedlib.enrage[aura.spellId] or hedlib.enrage[aura.name] then
				aura.dispelName="Enrage"
			end
		end
		if cfg.dispell[unitID].dispelName[aura.dispelName] then
			cfg.dispell[unitID].now=true
			cfg.dispell[unitID].instances=cfg.dispell[unitID].instances or {}
			if not cfg.dispell[unitID].instances[aura.auraInstanceID] then
				cfg.dispell[unitID].instances[aura.auraInstanceID]=true
				cfg.dispell[unitID].count=cfg.dispell[unitID].count and (cfg.dispell[unitID].count+1) or 1
				cfg.Update=true
			end
			aura_debug(aura,"Dispel aura Found",unitID)
			--print(GetTime().." Dispel aura Found on "..UnitName(unitID).."("..unitID..")"..aura.name.."("..aura.spellId..") "..aura.auraInstanceID)
			return true
		end
	end
	return false
end

local function RemoveDispelAura(auraInstanceID,unitID)
	unitID=unitID or cfg.aura_unit
	if cfg.dispell[unitID] and cfg.dispell[unitID].instances[auraInstanceID] then
		cfg.dispell[unitID].instances[auraInstanceID]=nil
		cfg.dispell[unitID].count=cfg.dispell[unitID].count-1
		if cfg.dispell[unitID].count==0 then
			cfg.dispell[unitID].now=false
		end
		aura_debug(nil,"Dispel aura Remove",unitID," "..auraInstanceID)
		cfg.Update=true
		return true
	end
	return false
end

local function ResetDispelAura(unitID)
	unitID=unitID or cfg.aura_unit
	if cfg.dispell[unitID] then
		if cfg.dispell[unitID].now then cfg.Update=true end
		aura_debug(nil,"Dispel aura Reset all",unitID)
		cfg.dispell[unitID].now=false
		cfg.dispell[unitID].instances={}
		cfg.dispell[unitID].count=0
	end
end


local function UpdateAura(aura,unitID)
	if not aura then return end
	unitID=unitID or cfg.aura_unit
	if CheckDispelAura(aura,unitID) then
		return
	end

	if cfg.aura_unit_id[unitID] and cfg.aura_unit_id[unitID][aura.spellId] 
	and (cfg.aura_unit_id[unitID][aura.spellId].sourceUnit==aura.sourceUnit
	or cfg.aura_unit_id[unitID][aura.spellId].sourceUnit=="any") then
		cfg.aura_unit_instance[unitID][aura.auraInstanceID]=cfg.aura_unit_id[unitID][aura.spellId]
		if cfg.aura_unit_id[unitID][aura.spellId].instance then
			cfg.aura_unit_id[unitID][aura.spellId].instance=aura
			lib.RunAuraFunction(aura.spellId,"OnUpdate")
			aura_debug(aura,"Aura Update",unitID)
		else
			cfg.aura_unit_id[unitID][aura.spellId].instance=aura
			lib.RunAuraFunction(aura.spellId,"OnUpdate")
			lib.RunAuraFunction(aura.spellId,"OnApply")
			lib.RunAuraFunction(aura.spellId,"OnStacks")
			aura_debug(aura,"Aura Apply",unitID)
		end
		cfg.Update=true
	end
end

local function RemoveAura(auraInstanceID,unitID)
	unitID=unitID or cfg.aura_unit
	aura_debug(nil,"removing",unitID," "..auraInstanceID)
	if RemoveDispelAura(auraInstanceID,unitID) then
		return
	end
	if cfg.aura_unit_instance[unitID][auraInstanceID] and cfg.aura_unit_instance[unitID][auraInstanceID].instance then
		if cfg.aura_unit_instance[unitID][auraInstanceID].instance.auraInstanceID==auraInstanceID then
			spellId=cfg.aura_unit_instance[unitID][auraInstanceID].instance.spellId
			aura_debug(cfg.aura_unit_instance[unitID][auraInstanceID].instance,"Aura Removed",unitID," "..auraInstanceID)
			cfg.aura_unit_instance[unitID][auraInstanceID].instance=nil
			cfg.aura_unit_instance[unitID][auraInstanceID]=nil
			lib.RunAuraFunction(spellId,"OnUpdate")
			lib.RunAuraFunction(spellId,"OnRemove")
			cfg.Update=true
		else
			cfg.aura_unit_instance[unitID][auraInstanceID]=nil
		end
	end
	
end

lib.ResetAuraAll=function(unitID)
	if not cfg.watchunits[unitID] then return end
	aura_debug(nil,"ResetAll",unitID)
	ResetDispelAura(unitID)
	if cfg.aura_unit_instance[unitID] then
		for auraInstanceID,_ in pairs(cfg.aura_unit_instance[unitID]) do
			RemoveAura(auraInstanceID,unitID)
		end
	end
end


lib.UpdateAuraAll = function(unitID)
	if not cfg.watchunits[unitID] then return end
	aura_debug(nil,"UpdateAll",unitID)
	--ResetDispelAura(unitID)
	--[[if not UnitExists(unitID) then
		lib.ResetAura(unitID)
		return
	end]]
	local batchCount = nil;
	local usePackedAura = true;
	cfg.aura_unit=unitID
	AuraUtil.ForEachAura(unitID, "HELPFUL", batchCount, UpdateAura, usePackedAura)
	AuraUtil.ForEachAura(unitID, "HARMFUL", batchCount, UpdateAura, usePackedAura)
end

function lib.UpdateAurasIncremental(unitID,unitAuraUpdateInfo)
	if not cfg.watchunits[unitID] then return end
	--cfg.aura_unit=unitID
	--aura_debug(nil,"UpdateAurasIncremental",unitID)
	if unitAuraUpdateInfo.addedAuras ~= nil then
		aura_debug(nil,"addedAuras",unitID)
        for _, aura in ipairs(unitAuraUpdateInfo.addedAuras) do
			UpdateAura(aura,unitID)
        end
    end

    if unitAuraUpdateInfo.updatedAuraInstanceIDs ~= nil then
		aura_debug(nil,"updatedAuraInstanceIDs",unitID)
        for _, auraInstanceID in ipairs(unitAuraUpdateInfo.updatedAuraInstanceIDs) do
			UpdateAura(C_UnitAuras.GetAuraDataByAuraInstanceID(unitID, auraInstanceID),unitID)
        end
    end

    if unitAuraUpdateInfo.removedAuraInstanceIDs ~= nil then
		aura_debug(nil,"removedAuraInstanceIDs",unitID)
        for _, auraInstanceID in ipairs(unitAuraUpdateInfo.removedAuraInstanceIDs) do
			RemoveAura(auraInstanceID,unitID)
			--RemoveDispelAura(auraInstanceID,unitID)
			
        end
    end
end


--[[lib.AuraInTime = function (spell , mint, maxt)
	if not cfg.aura[spell] then return nil end
	if cfg.aura[spell].faded >= mint and cfg.aura[spell].faded <=maxt then
		return true
	else
		return nil
	end
end]]

--[[lib.GetAuraLastTime = function (spell)
	if cfg.aura[spell] then
		return cfg.aura[spell].faded
	end
	return 0
end

lib.GetAuraLastTimeStacks = function (spell)
	if cfg.aura[spell] then
		return cfg.aura[spell].pstacks
	end
	return 0
end]]


local tl_aura,tl_caura
-- options = any,min,max

local function Aura(spellId)
	if cfg.aura[spellId] and cfg.aura[spellId].instance then
		return (cfg.aura[spellId].instance.duration==0 and 9999 or math.max(0,cfg.aura[spellId].instance.expirationTime-GetTime()))
	end
	return 0
end

lib.GetAura = function(auras,option)
	option=option or "first"
	tl_aura = 0
	if type(auras)~="table" then
		auras={auras}
	end
	for _,spellId in ipairs(auras) do
		tl_caura = Aura(spellId)
		if tl_caura>0 then
			if option=="max" then
				tl_aura=math.max(tl_aura,tl_caura)
			else
				return tl_caura
			end
		end
	end
	return tl_aura
end

lib.GetAuraCaster = function(aura,option)
	if lib.GetAura({aura},option)>0 then
		return cfg.aura[aura].real_caster
	else
		return nil
	end
end

local i_aura
--[[lib.GetAuras = function(aura)
	tl_aura = 0
	i_aura=1
	notfound_aura=true
	while notfound_aura do
		n_aura=aura..i_aura
		if cfg.aura[n_aura] then
			if cfg.aura[n_aura].present==1 then
				tl_aura = (cfg.aura[n_aura].duration==0 and 9999 or math.max(0,cfg.aura[n_aura].expire-GetTime())
				if tl_aura>0 then
					notfound_aura=false
					return tl_aura
				end
			end
			i_aura=i_aura+1
		else
			notfound_aura=false
			return 0
		end
	end
	return 0
end]]

local auras={}
lib.GetAuras = function(aura,option)
	option = option or "max"
	if cfg.auras[aura] then
		return lib.GetAura(cfg.auras[aura],option)
	else
		return 0
	end
end


lib.GetAuraIcon = function(aura)
	if cfg.aura[aura] then return cfg.aura[aura].icon end
	icon=(select(3,GetSpellInfo(aura)))
	return icon
end

lib.HasAuras = function(aura)
	i_aura=1
	notfound_aura=true
	while notfound_aura do
		n_aura=aura..i_aura
		if cfg.aura[n_aura] then
			if cfg.aura[n_aura].present==1 then
				return true
			end
			i_aura=i_aura+1
		else
			notfound_aura=nil
		end
	end
	return nil
end

lib.IsAura = function(aura,cast)
	if cfg.aura[aura] then
		if cfg.aura[aura].present~=0 or (cast and GetTime()-cfg.aura[aura].faded<cfg.gcd) then
			return true
		end
	end
	return false
end

-- lib.GetShortestAura = function(auras)
	-- tl_aura=99999
	-- for _,n_aura in ipairs(auras) do
		-- if cfg.aura[n_aura] then
			-- tl_caura=cfg.aura[n_aura].expire-GetTime()
			-- tl_caura = (tl_caura>=0) and tl_caura or 0
			-- if tl_caura<tl_aura then tl_aura=tl_caura end
		-- end
	-- end
	-- if tl_aura==99999 then return 0 end
	-- return tl_aura
-- end

-- lib.GetLargestAura = function(auras)
	-- tl_aura=0
	-- for _,n_aura in ipairs(auras) do
		-- if cfg.aura[n_aura] then
			-- tl_caura=cfg.aura[n_aura].expire-GetTime()
			-- tl_caura = (tl_caura>=0) and tl_caura or 0
			-- if tl_caura>tl_aura then tl_aura=tl_caura end
		-- end
	-- end
	-- return tl_aura
-- end

-- lib.GetLargestNumber = function(numbers)
	-- bignumber=0
	-- for _,n_num in ipairs(numbers) do
		-- if n_num>bignumber then bignumber=n_num end
	-- end
	-- return bignumber
-- end

local tl_aura
lib.GetAuraStacks = function(spellId)
	if cfg.aura[spellId] and cfg.aura[spellId].instance then
		if cfg.aura[spellId].instance.applications then return cfg.aura[spellId].instance.applications end
		return 1
	end
	return 0
end

--[[lib.UpdateWeaponBuffs = function()
	pMain, pOff = cfg.Weapons.Main, cfg.Weapons.Off
	cfg.Weapons.Main, _, _, cfg.Weapons.Off = GetWeaponEnchantInfo()
	if pMain~=cfg.Weapons.Main or pOff~=cfg.Weapons.Off then return true end
	return false
end]]



--[[lib.PrintAura = function(spell)
	if cfg.aura[spell] then
		tmp=cfg.aura[spell].LocalName
		tmp=tmp.." ("..cfg.aura[spell].spellId..") "
		tmp=tmp..cfg.aura[spell].powerType
		tmp=tmp.." on "..cfg.aura[spell].unit
		tmp=tmp.." caster "..cfg.aura[spell].caster
		tmp=tmp.." stacks="..cfg.aura[spell].stacks
		tmp=tmp.." left="..(cfg.aura[spell].duration==0 and 9999 or math.max(0,cfg.aura[spell].expire-GetTime()))
		tmp=tmp.." start="..(cfg.aura[spell].expire-cfg.aura[spell].duration)
		tmp=tmp.." duration="..cfg.aura[spell].duration
		tmp=tmp.." expire="..cfg.aura[spell].expire
		print(tmp)
	end
end]]


--[[lib.PrintAuras = function(aura)
	local n_aura
	print(aura.." "..lib.GetAuras(aura))
	for _,n_aura in ipairs(cfg.auras[aura]) do
		print(n_aura.." - "..cfg.aura[n_aura].spellId.." "..cfg.aura[n_aura].LocalName.." "..lib.GetAura({n_aura}))
	end
end]]

--[[lib.PrintBuff = function(unit, bufftype)
	unit=unit or "player"
	bufftype=bufftype or "all"
	for i=1,40 do
		local name, icon, count, b_type, duration, b_exp, caster, stealable, _, SpellID=UnitAura(unit,i)
		if name then
			if count==0 then count=1 end
			b_type=b_type or "Other"
			if b_type=="" and stealable then b_type="Enrage" end
			if hedlib.enrage[SpellID] then b_type="Enrage" end
			if b_type==bufftype or bufftype=="all" then
				tmp=name
				tmp=tmp.." ("..SpellID..") "
				tmp=tmp..b_type
				tmp=tmp.." count="..count
				tmp=tmp.." left="..(duration==0 and 9999 or math.max(0,b_exp-GetTime()))
				tmp=tmp.." duration="..duration
				tmp=tmp.." expire="..b_exp
				tmp=tmp.." caster="..(caster or "unknown")
				print(tmp)
			end
		end
	end
end

lib.PrintDebuff = function(unit, bufftype)
	unit=unit or "target"
	bufftype=bufftype or "all"
	for i=1,40 do
		local name, icon, count, b_type, duration, b_exp, caster, stealable, _, SpellID=UnitDebuff(unit,i)
		if name then
			if count==0 then count=1 end
			b_type=b_type or "Other"
			if hedlib.enrage[SpellID] then b_type="Enrage" end
			if b_type==bufftype or bufftype=="all" then
				tmp=name
				tmp=tmp.." ("..SpellID..") "
				tmp=tmp..b_type
				tmp=tmp.." count="..count
				tmp=tmp.." left="..(duration==0 and 9999 or math.max(0,b_exp-GetTime()))
				tmp=tmp.." duration="..duration
				tmp=tmp.." expire="..b_exp
				tmp=tmp.." caster="..(caster or "unknown")
				print(tmp)
			end
		end
	end
end]]





local shortestaura
local shortestaurac
local shortestauraname
lib.GetAuraShortestName = function(auras)
	shortestaura=99999
	for k,spell in pairs(auras) do
		shortestaurac=lib.GetAura(spell)
		if shortestaurac<shortestaura then
			shortestaura=shortestaurac
			shortestauraname=spell
		end
	end
	return shortestauraname
end


--good
lib.GetAuraApplyTime = function(spell)
	if cfg.aura[spell] and cfg.aura[spell].instance then
		--if cfg.aura[spell].instance.==0 then return 0 end
		return (cfg.aura[spell].instance.expirationTime-cfg.aura[spell].instance.duration)
	end
	return 0
end

lib.GetAuraDuration = function(spell)
	if cfg.aura[spell] and cfg.aura[spell].instance then
		return cfg.aura[spell].instance.duration
	end
	return 0
end

lib.GetAuraId = function (spell)
	if cfg.aura[spell] then
		return cfg.aura[spell].spellId
	end
	return "noaura"
end

lib.GetAuraRefresh = function(spell)
	if cfg.aura[spell] then
		return cfg.aura[spell].refresh or 0
	end
	return 0
end

lib.SpellRefreshAura = function(spell,aura)
	aura = aura or spell
	--if (not aura) then aura=spell end
	return lib.SimpleCDCheck(spell, lib.GetAura({aura})-lib.GetAuraRefresh(aura))
end

lib.SpellApplyAura = function(spell,aura)
	aura = aura or spell
	--if (not aura) then aura=spell end
	return lib.SimpleCDCheck(spell, lib.GetAura({aura}))
end

lib.SetAuraRefresh = function(spell,refresh)
	if cfg.aura[spell] then
		cfg.aura[spell].refresh=refresh or 0
	end
end

lib.AddDispell = function(spell,SpellIDs,types,unit)
	lib.AddSpell(spell,SpellIDs)
	if cfg.spells[spell] then
		cfg.watchunits[unit]=true
		cfg.dispell[unit]={}
		cfg.dispell[unit].now=false
		cfg.dispell[unit].dispelName={}
		for _,n_type in ipairs(types) do
			cfg.dispell[unit].dispelName[n_type]=true
		end
		cfg.dispell[unit].Name=spell
		cfg.case[spell] = function()
			if cfg.dispell[unit].now then
				return lib.SimpleCDCheck(spell)
			end
			return nil
		end
	end
end